"""This script tests the cli methods to add families to status-db"""
from datetime import datetime, timedelta

from cg.constants import CASE_ACTIONS, PRIORITY_OPTIONS, DataDelivery, Pipeline
from cg.store import Store


def test_delivered_at_affects_tat(base_store: Store, helpers):
    """test that the estimated turnaround time is affected by the delivered_at date """

    # GIVEN a database with a case and a samples receive_at, prepared_at, sequenced_at,
    # delivered_at one week ago
    new_case = add_case(helpers, base_store, ordered_days_ago=7)
    one_week_ago = datetime.now() - timedelta(days=7)
    one_week_old_sample = helpers.add_sample(
        base_store,
        ordered_at=one_week_ago,
        received_at=one_week_ago,
        prepared_at=one_week_ago,
        sequenced_at=one_week_ago,
        delivered_at=one_week_ago,
    )
    base_store.relate_sample(new_case, one_week_old_sample, "unknown")

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN TAT should be R-D = 0 = 0
    assert cases
    for case in cases:
        assert case.get("tat") == 0


def test_sequenced_at_affects_tat(base_store: Store, helpers):
    """test that the estimated turnaround time is affected by the sequenced_at date """

    # GIVEN a database with a case and a samples receive_at, prepared_at, sequenced_at one week
    # ago
    new_case = add_case(helpers, base_store, ordered_days_ago=7)
    one_week_ago = datetime.now() - timedelta(days=7)
    one_week_old_sample = helpers.add_sample(
        base_store,
        ordered_at=one_week_ago,
        received_at=one_week_ago,
        prepared_at=one_week_ago,
        sequenced_at=one_week_ago,
    )
    base_store.relate_sample(new_case, one_week_old_sample, "unknown")

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN TAT should be R-P + P-S + S-A + A-U + U-D = 0 + 0 + 7 + 1 + 2 = 10
    assert cases
    for case in cases:
        assert case.get("tat") == 10


def test_prepared_at_affects_tat(base_store: Store, helpers):
    """test that the estimated turnaround time is affected by the prepared_at date """

    # GIVEN a database with a case and a samples receive_at, prepared_at one week ago
    new_case = add_case(helpers, base_store, ordered_days_ago=7)
    one_week_ago = datetime.now() - timedelta(days=7)
    one_week_old_sample = helpers.add_sample(
        base_store,
        ordered_at=one_week_ago,
        received_at=one_week_ago,
        prepared_at=one_week_ago,
    )
    base_store.relate_sample(new_case, one_week_old_sample, "unknown")

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN TAT should be R-P + P-S + S-A + A-U + U-D = 0 + 7 + 4 + 1 + 2 = 14
    assert cases
    for case in cases:
        assert case.get("tat") == 14


def test_received_at_affects_tat(base_store: Store, helpers):
    """test that the estimated turnaround time is affected by the received_at date """

    # GIVEN a database with a case and a samples received one week ago
    new_case = add_case(helpers, base_store, ordered_days_ago=7)
    one_week_ago = datetime.now() - timedelta(days=7)
    one_week_old_sample = helpers.add_sample(
        base_store,
        ordered_at=one_week_ago,
        received_at=one_week_ago,
    )
    base_store.relate_sample(new_case, one_week_old_sample, "unknown")

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN TAT should be R-P + P-S + S-A + A-U + U-D = 7 + 5 + 4 + 1 + 2 = 19
    assert cases
    for case in cases:
        assert case.get("tat") == 19


def test_samples_flowcell(base_store: Store, helpers):
    """Test to that cases displays the flowcell status """

    # GIVEN a database with a case with a sample that belongs to a flowcell with status ondisk
    # and a sample not yet on a flowcell
    new_case = add_case(helpers, base_store)
    sample_on_flowcell = helpers.add_sample(base_store)
    flowcell = helpers.add_flowcell(base_store, status="ondisk", samples=[sample_on_flowcell])
    base_store.relate_sample(new_case, sample_on_flowcell, "unknown")
    sample_not_on_flowcell = helpers.add_sample(base_store)
    base_store.relate_sample(new_case, sample_not_on_flowcell, "unknown")

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain info on flowcell status and be false
    assert cases
    for case in cases:
        assert flowcell.status in case.get("flowcells_status")
        assert not case.get("flowcells_on_disk_bool")
        assert case.get("flowcells_on_disk") == 1


def test_sample_flowcell(base_store: Store, helpers):
    """Test to that cases displays the flowcell status """

    # GIVEN a database with a case with a sample that belongs to a flowcell with status ondisk
    new_case = add_case(helpers, base_store)
    sample = helpers.add_sample(base_store)
    base_store.relate_sample(new_case, sample, "unknown")
    flowcell = helpers.add_flowcell(base_store, status="ondisk", samples=[sample])
    assert flowcell.status

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain info on flowcell status and be true
    assert cases
    for case in cases:
        assert case.get("flowcells_on_disk") == 1
        assert case.get("flowcells_status") == flowcell.status
        assert case.get("flowcells_on_disk_bool")


def test_case_action(base_store: Store, helpers):
    """Test that case action is fetched by cases() method """

    # GIVEN a database with an analysis that was completed but has an active rerun in progress
    analysis = helpers.add_analysis(
        base_store, completed_at=datetime.now(), uploaded_at=datetime.now()
    )
    analysis.family.action = "analyze"

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain info on analysis (case) action
    assert cases
    for case in cases:
        assert case.get("case_action") == analysis.family.action


def test_analysis_dates_for_rerun(base_store: Store, helpers):
    """Test to that cases displays no analysis dates for active reruns """

    # GIVEN a database with an analysis that was completed but has an active rerun in progress
    analysis = helpers.add_analysis(
        base_store, completed_at=datetime.now(), uploaded_at=datetime.now()
    )
    analysis.family.action = "analyze"

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should not contain info on completed and uploaded occasion
    assert cases
    for case in cases:
        assert case.get("analysis_completed_at") is None
        assert case.get("analysis_uploaded_at") is None


def test_received_at_is_newest_date(base_store: Store, helpers):
    """Test to that cases displays newest received date"""

    # GIVEN a database with a case and two samples with different received dates
    new_case = add_case(helpers, base_store)
    yesterday = datetime.now() - timedelta(days=1)
    yesteryear = datetime.now() - timedelta(days=365)
    newest_sample = helpers.add_sample(base_store, received_at=yesterday)
    oldest_sample = helpers.add_sample(base_store, received_at=yesteryear)
    base_store.relate_sample(new_case, newest_sample, "unknown")
    base_store.relate_sample(new_case, oldest_sample, "unknown")

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain the date the samples were most recently received
    assert cases
    for case in cases:
        assert case.get("samples_received_at").date() == newest_sample.received_at.date()


def test_prepared_at_is_newest_date(base_store: Store, helpers):
    """Test to that cases displays newest prepared date"""

    # GIVEN a database with a case and two samples with different prepared dates
    new_case = add_case(helpers, base_store)
    yesterday = datetime.now() - timedelta(days=1)
    yesteryear = datetime.now() - timedelta(days=365)
    newest_sample = helpers.add_sample(base_store, prepared_at=yesterday)
    oldest_sample = helpers.add_sample(base_store, prepared_at=yesteryear)
    base_store.relate_sample(new_case, newest_sample, "unknown")
    base_store.relate_sample(new_case, oldest_sample, "unknown")

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain the date the samples were most recently prepared
    assert cases
    for case in cases:
        assert case.get("samples_prepared_at").date() == newest_sample.prepared_at.date()


def test_sequenced_at_is_newest_date(base_store: Store, helpers):
    """Test to that cases displays newest sequenced date"""

    # GIVEN a database with a case and two samples with different sequenced dates
    new_case = add_case(helpers, base_store)
    yesterday = datetime.now() - timedelta(days=1)
    yesteryear = datetime.now() - timedelta(days=365)
    newest_sample = helpers.add_sample(base_store, sequenced_at=yesterday)
    oldest_sample = helpers.add_sample(base_store, sequenced_at=yesteryear)
    base_store.relate_sample(new_case, newest_sample, "unknown")
    base_store.relate_sample(new_case, oldest_sample, "unknown")

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain the date the samples were most recently sequenced
    assert cases
    for case in cases:
        assert case.get("samples_sequenced_at").date() == newest_sample.sequenced_at.date()


def test_delivered_at_is_newest_date(base_store: Store, helpers):
    """Test to that cases displays newest delivered date"""

    # GIVEN a database with a case and two samples with different delivered dates
    new_case = add_case(helpers, base_store)
    yesterday = datetime.now() - timedelta(days=1)
    yesteryear = datetime.now() - timedelta(days=365)
    newest_sample = helpers.add_sample(base_store, delivered_at=yesterday)
    oldest_sample = helpers.add_sample(base_store, delivered_at=yesteryear)
    base_store.relate_sample(new_case, newest_sample, "unknown")
    base_store.relate_sample(new_case, oldest_sample, "unknown")

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain the date the samples were most recently delivered
    assert cases
    for case in cases:
        assert case.get("samples_delivered_at").date() == newest_sample.delivered_at.date()


def test_invoiced_at_is_newest_invoice_date(base_store: Store, helpers):
    """Test to that cases displays newest invoiced date"""

    # GIVEN a database with a case and two samples with different invoiced dates
    new_case = add_case(helpers, base_store)
    yesterday = datetime.now() - timedelta(days=1)
    yesteryear = datetime.now() - timedelta(days=365)
    newest_sample = helpers.add_sample(base_store)
    invoice = base_store.add_invoice(helpers.ensure_customer(base_store))
    newest_sample.invoice = invoice
    newest_sample.invoice.invoiced_at = yesterday
    oldest_sample = helpers.add_sample(base_store)
    invoice = base_store.add_invoice(helpers.ensure_customer(base_store))
    oldest_sample.invoice = invoice
    oldest_sample.invoice.invoiced_at = yesteryear
    base_store.relate_sample(new_case, newest_sample, "unknown")
    base_store.relate_sample(new_case, oldest_sample, "unknown")

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain the date the samples were most recently invoiced
    assert cases
    for case in cases:
        assert case.get("samples_invoiced_at").date() == newest_sample.invoice.invoiced_at.date()


def test_invoiced_at(base_store: Store, helpers):
    """Test to that cases displays correct invoiced date"""

    # GIVEN a database with a case and a invoiced date
    new_case = add_case(helpers, base_store)
    sample = helpers.add_sample(base_store)
    sample.invoice = base_store.add_invoice(helpers.ensure_customer(base_store))
    sample.invoice.invoiced_at = datetime.now()
    base_store.relate_sample(new_case, sample, "unknown")

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain the date the sample was invoiced
    assert cases
    for case in cases:
        assert case.get("samples_invoiced_at").date() == datetime.now().date()


def test_delivered_at(base_store: Store, helpers):
    """Test to that cases displays correct delivered date"""

    # GIVEN a database with a case and a delivered date
    new_case = add_case(helpers, base_store)
    sample = helpers.add_sample(base_store, delivered_at=datetime.now())
    base_store.relate_sample(new_case, sample, "unknown")

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain the date the sample was delivered
    assert cases
    for case in cases:
        assert case.get("samples_delivered_at").date() == datetime.now().date()


def test_sequenced_at(base_store: Store, helpers):
    """Test to that cases displays correct sequenced date"""

    # GIVEN a database with a case and a sequenced date
    new_case = add_case(helpers, base_store)
    sample = helpers.add_sample(base_store, sequenced_at=datetime.now())
    base_store.relate_sample(new_case, sample, "unknown")

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain the date the sample was sequenced
    assert cases
    for case in cases:
        assert case.get("samples_sequenced_at").date() == datetime.now().date()


def test_prepared_at(base_store: Store, helpers):
    """Test to that cases displays correct prepared date"""

    # GIVEN a database with a case and a prepared date
    new_case = add_case(helpers, base_store)
    sample = helpers.add_sample(base_store, prepared_at=datetime.now())
    base_store.relate_sample(new_case, sample, "unknown")

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain the date the sample was prepared
    assert cases
    for case in cases:
        assert case.get("samples_prepared_at").date() == datetime.now().date()


def test_received_at(base_store: Store, helpers):
    """Test to that cases displays correct received date"""

    # GIVEN a database with a case and a received date
    new_case = add_case(helpers, base_store)
    sample = helpers.add_sample(base_store, received_at=datetime.now())
    base_store.relate_sample(new_case, sample, "unknown")

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain the date the sample was received
    assert cases
    for case in cases:
        assert case.get("samples_received_at").date() == datetime.now().date()


def test_no_invoice_true(base_store: Store, helpers):
    """Test to that cases displays correct samples to invoice"""

    # GIVEN a database with a case and one no_invoice sample
    new_case = add_case(helpers, base_store)
    sample = helpers.add_sample(base_store, no_invoice=True)
    base_store.relate_sample(new_case, sample, "unknown")
    assert sample.no_invoice

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain zero samples to invoice
    assert cases
    for case in cases:
        assert case.get("samples_to_invoice") == 0


def test_one_no_invoice_false(base_store: Store, helpers):
    """Test to that cases displays correct samples to invoice"""

    # GIVEN a database with a case and one sample to invoice
    new_case = add_case(helpers, base_store)
    sample = helpers.add_sample(base_store, no_invoice=False)
    base_store.relate_sample(new_case, sample, "unknown")
    assert not sample.no_invoice

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain one sample to invoice
    assert cases
    for case in cases:
        assert case.get("samples_to_invoice") == 1


def test_one_external_sample(base_store: Store, helpers):
    """Test to that cases displays correct internal/external samples"""

    # GIVEN a database with a case and one external sample
    new_case = add_case(helpers, base_store)
    sample = helpers.add_sample(base_store, is_external=True)
    base_store.relate_sample(new_case, sample, "unknown")
    assert sample.application_version.application.is_external

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain one external and zero internal samples
    assert cases
    for case in cases:
        assert case.get("total_external_samples") == 1
        assert case.get("total_internal_samples") == 0
        assert case.get("case_external_bool")
        assert case.get("samples_to_receive") == 0
        assert case.get("samples_to_prepare") == 0
        assert case.get("samples_to_sequence") == 0


def test_one_internal_sample(base_store: Store, helpers):
    """Test to that cases displays correct internal/external samples"""

    # GIVEN a database with a case and one external sample
    new_case = add_case(helpers, base_store)
    sample = helpers.add_sample(base_store, is_external=False)
    base_store.relate_sample(new_case, sample, "unknown")
    assert not sample.is_external

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain zero external and one internal samples
    assert cases
    for case in cases:
        assert case.get("total_external_samples") == 0
        assert case.get("total_internal_samples") == 1
        assert case.get("samples_to_receive") == 1
        assert case.get("samples_to_prepare") == 1
        assert case.get("samples_to_sequence") == 1


def test_include_case_by_sample_id(base_store: Store, helpers):
    """Test to that cases can be included by sample id"""

    # GIVEN a database with a sample with internal_id
    new_case = add_case(helpers, base_store)
    sample = helpers.add_sample(base_store)
    base_store.relate_sample(new_case, sample, "unknown")

    # WHEN getting active cases by sample id
    cases = base_store.cases(sample_id=sample.internal_id)

    # THEN cases should only contain this case
    assert cases
    for case in cases:
        assert new_case.internal_id in case.get("internal_id")


def test_exclude_case_by_sample_id(base_store: Store, helpers):
    """Test to that cases can be excluded by sample id"""

    # GIVEN a database with a sample with internal_id
    new_case = add_case(helpers, base_store)
    sample = helpers.add_sample(base_store)
    base_store.relate_sample(new_case, sample, "unknown")

    # WHEN getting active cases by non-existing sample id
    cases = base_store.cases(sample_id="dummy_id")

    # THEN cases should not contain this case
    assert not cases


def test_include_case_by_exclude_customer(base_store: Store, helpers):
    """Test to that cases can be excluded by customer"""

    # GIVEN a database with a case and a customer
    customer_id = "cust000"
    new_case = add_case(helpers, base_store, customer_id=customer_id)

    # WHEN getting active cases by customer
    cases = base_store.cases(exclude_customer_id="dummy_customer")

    # THEN cases should contain this case
    assert cases
    for case in cases:
        assert new_case.internal_id in case.get("internal_id")


def test_exclude_case_by_exclude_customer(base_store: Store, helpers):
    """Test to that cases can be excluded by customer"""

    # GIVEN a database with a case and a customer
    customer_id = "cust000"
    add_case(helpers, base_store, customer_id=customer_id)

    # WHEN getting active cases by customer
    cases = base_store.cases(exclude_customer_id=customer_id)

    # THEN cases should not contain this case
    assert not cases


def test_include_case_by_case_uppercase_data_analysis(base_store: Store, helpers):
    """Test to that cases can be included by uppercase data_analysis"""

    # GIVEN a database with a case with data analysis set
    data_analysis = Pipeline.BALSAMIC
    assert str(data_analysis).upper() != str(data_analysis)
    new_case = add_case(helpers, base_store, data_analysis=data_analysis)
    sample = helpers.add_sample(base_store)
    base_store.relate_sample(new_case, sample, "unknown")

    # WHEN getting active cases by data_analysis
    cases = base_store.cases(data_analysis=new_case.data_analysis.upper())

    # THEN cases should only contain this case
    assert cases
    for case in cases:
        assert new_case.internal_id in case.get("internal_id")


def test_exclude_case_by_data_analysis(base_store: Store, helpers):
    """Test to that cases can be excluded by data_analysis"""

    # GIVEN a database with a case with data analysis set
    new_case = add_case(helpers, base_store, data_analysis=Pipeline.BALSAMIC)
    sample = helpers.add_sample(base_store)
    base_store.relate_sample(new_case, sample, "unknown")

    # WHEN getting active cases by data_analysis
    cases = base_store.cases(data_analysis="dummy_analysis")

    # THEN cases should not contain this case
    assert not cases


def test_include_case_by_partial_data_analysis(base_store: Store, helpers):
    """Test to that cases can be included by data_analysis"""

    # GIVEN a database with a case with data analysis set
    data_analysis = Pipeline.BALSAMIC
    new_case = add_case(helpers, base_store, data_analysis=data_analysis)
    sample = helpers.add_sample(base_store)
    base_store.relate_sample(new_case, sample, "unknown")

    # WHEN getting active cases by partial data_analysis
    cases = base_store.cases(data_analysis=str(data_analysis)[:-1])

    # THEN cases should only contain this case
    assert cases
    for case in cases:
        assert new_case.internal_id in case.get("internal_id")


def test_show_multiple_data_analysis(base_store: Store, helpers):
    """Test to that cases can be included by data_analysis"""

    # GIVEN a database with a case with data analysis set
    data_analysis = Pipeline.BALSAMIC
    new_case = add_case(helpers, base_store, data_analysis=data_analysis)
    sample1 = helpers.add_sample(base_store)
    base_store.relate_sample(new_case, sample1, "unknown")
    new_case2 = add_case(helpers, base_store, case_id="new_case2", data_analysis=data_analysis)
    sample2 = helpers.add_sample(base_store)
    base_store.relate_sample(new_case, sample2, "unknown")
    base_store.relate_sample(new_case2, sample2, "unknown")

    # WHEN getting active cases by data_analysis
    cases = base_store.cases(data_analysis=str(data_analysis)[:-1])

    # THEN cases should only contain these analyses
    assert cases
    for case in cases:
        assert new_case2.internal_id or new_case.internal_id in case.get("internal_id")


def test_show_data_analysis(base_store: Store, helpers):
    """Test to that cases can be included by data_analysis"""

    # GIVEN a database with a case with data analysis set
    data_analysis = Pipeline.BALSAMIC
    new_case = add_case(helpers, base_store, data_analysis=data_analysis)
    sample = helpers.add_sample(base_store)
    base_store.relate_sample(new_case, sample, "unknown")

    # WHEN getting active cases by data_analysis
    cases = base_store.cases(data_analysis=str(data_analysis))

    # THEN cases should only contain this case
    assert cases
    for case in cases:
        assert new_case.internal_id in case.get("internal_id")


def test_include_case_by_data_analysis(base_store: Store, helpers):
    """Test to that cases can be included by data_analysis"""

    # GIVEN a database with a case with data analysis set
    data_analysis = Pipeline.BALSAMIC
    new_case = add_case(helpers, base_store, data_analysis=data_analysis)
    sample = helpers.add_sample(base_store)
    base_store.relate_sample(new_case, sample, "unknown")

    # WHEN getting active cases by data_analysis
    cases = base_store.cases(data_analysis=new_case.data_analysis)

    # THEN cases should only contain this case
    assert cases
    for case in cases:
        assert new_case.internal_id in case.get("internal_id")


def test_exclude_case_by_customer(base_store: Store, helpers):
    """Test to that cases can be excluded by customer"""

    # GIVEN a database with a case and a customer
    add_case(helpers, base_store, customer_id="cust000")

    # WHEN getting active cases by customer
    cases = base_store.cases(customer_id="dummy_cust")

    # THEN cases should not contain this case
    assert not cases


def test_include_case_by_customer(base_store: Store, helpers):
    """Test to that cases can be included by customer"""

    # GIVEN a database with a case
    customer_id = "cust000"
    new_case = add_case(helpers, base_store, customer_id=customer_id)

    # WHEN getting active cases by customer
    cases = base_store.cases(customer_id=customer_id)

    # THEN cases should only contain this case
    assert cases
    for case in cases:
        assert new_case.internal_id in case.get("internal_id")


def test_exclude_case_by_name(base_store: Store, helpers):
    """Test to that cases can be excluded by name"""

    # GIVEN a database with a case
    add_case(helpers, base_store)

    # WHEN getting active cases by name
    cases = base_store.cases(name="dummy_name")

    # THEN cases should not contain this case
    assert not cases


def test_include_case_by_partial_name(base_store: Store, helpers):
    """Test to that cases can be included by name"""

    # GIVEN a database with a case
    new_case = add_case(helpers, base_store)

    # WHEN getting active cases by partial name
    cases = base_store.cases(name=new_case.name[1:-1])

    # THEN cases should only contain this case
    assert cases
    for case in cases:
        assert new_case.name in case.get("name")


def test_include_case_by_name(base_store: Store, helpers):
    """Test to that cases can be included by name"""

    # GIVEN a database with a case
    new_case = add_case(helpers, base_store)

    # WHEN getting active cases by name
    cases = base_store.cases(name=new_case.name)

    # THEN cases should only contain this case
    assert cases
    for case in cases:
        assert new_case.name in case.get("name")


def test_excluded_by_priority(base_store: Store, helpers):
    """Test to that cases can be excluded by priority"""

    # GIVEN a database with a case with a priority
    add_case(helpers, base_store, priority=PRIORITY_OPTIONS[0])

    # WHEN getting active cases by priority
    cases = base_store.cases(priority=PRIORITY_OPTIONS[1])

    # THEN cases should not contain this case
    assert not cases


def test_included_by_priority(base_store: Store, helpers):
    """Test to that cases can be included by priority"""

    # GIVEN a database with a case with a priority
    new_case = add_case(helpers, base_store, priority=PRIORITY_OPTIONS[0])

    # WHEN getting active cases by priority
    cases = base_store.cases(priority=new_case.action)

    # THEN cases should only contain this case
    assert cases
    for case in cases:
        assert new_case.internal_id in case.get("internal_id")


def test_excluded_by_action(base_store: Store, helpers):
    """Test to that cases can be excluded by action"""

    # GIVEN a database with a case with an action
    add_case(helpers, base_store, action=CASE_ACTIONS[0])

    # WHEN getting active cases by action
    cases = base_store.cases(case_action=CASE_ACTIONS[1])

    # THEN cases should not contain this case
    assert not cases


def test_included_by_action(base_store: Store, helpers):
    """Test to that cases can be included by action"""

    # GIVEN a database with a case with an action
    new_case = add_case(helpers, base_store, action=CASE_ACTIONS[0])

    # WHEN getting active cases by action
    cases = base_store.cases(case_action=new_case.action)

    # THEN cases should only contain this case
    assert cases
    for case in cases:
        assert new_case.internal_id in case.get("internal_id")


def test_exclude_case_by_internal_id(base_store: Store, helpers):
    """Test to that cases can be excluded by internal_id"""

    # GIVEN a database with a case
    add_case(helpers, base_store)

    # WHEN getting active cases by internal_id
    cases = base_store.cases(internal_id="dummy_id")

    # THEN cases should not contain this case
    assert not cases


def test_include_case_by_partial_internal_id(base_store: Store, helpers):
    """Test to that cases can be included by internal_id"""

    # GIVEN a database with a case
    new_case = add_case(helpers, base_store)

    # WHEN getting active cases by partial internal_id
    cases = base_store.cases(internal_id=new_case.internal_id[1:-1])

    # THEN cases should only contain this case
    assert cases
    for case in cases:
        assert new_case.internal_id in case.get("internal_id")


def test_include_case_by_internal_id(base_store: Store, helpers):
    """Test to that cases can be included by internal_id"""

    # GIVEN a database with a case
    new_case = add_case(helpers, base_store)

    # WHEN getting active cases by internal_id
    cases = base_store.cases(internal_id=new_case.internal_id)

    # THEN cases should only contain this case
    assert cases
    for case in cases:
        assert new_case.internal_id in case.get("internal_id")


def test_only_prepared_cases(base_store: Store, helpers):
    """Test to that sequenced cases can be included"""

    # GIVEN a database with an prepared case
    new_case = add_case(helpers, base_store)
    sample = helpers.add_sample(base_store, prepared_at=datetime.now())
    base_store.relate_sample(new_case, sample, "unknown")
    neg_new_case = add_case(helpers, base_store, "neg_new_case")
    neg_sample = helpers.add_sample(base_store, sample_name="neg_sample")
    base_store.relate_sample(neg_new_case, neg_sample, "unknown")

    # WHEN getting active cases excluding prepared
    cases = base_store.cases(only_prepared=True)

    # THEN cases should only contain the prepared case
    assert cases
    for case in cases:
        assert new_case.internal_id in case.get("internal_id")


def test_only_received_cases(base_store: Store, helpers):
    """Test to that sequenced cases can be included"""

    # GIVEN a database with an received case
    new_case = add_case(helpers, base_store)
    sample = helpers.add_sample(base_store, received_at=datetime.now())
    base_store.relate_sample(new_case, sample, "unknown")
    neg_new_case = add_case(helpers, base_store, "neg_new_case")
    neg_sample = helpers.add_sample(base_store, sample_name="neg_sample")
    base_store.relate_sample(neg_new_case, neg_sample, "unknown")

    # WHEN getting active cases excluding received
    cases = base_store.cases(only_received=True)

    # THEN cases should only contain the received case
    assert cases
    for case in cases:
        assert new_case.internal_id in case.get("internal_id")


def test_only_sequenced_cases(base_store: Store, helpers):
    """Test to that sequenced cases can be included"""

    # GIVEN a database with an sequenced case
    new_case = add_case(helpers, base_store)
    sample = helpers.add_sample(base_store, sequenced_at=datetime.now())
    base_store.relate_sample(new_case, sample, "unknown")
    neg_new_case = add_case(helpers, base_store, "neg_new_case")
    neg_sample = helpers.add_sample(base_store, sample_name="neg_sample")
    base_store.relate_sample(neg_new_case, neg_sample, "unknown")

    # WHEN getting active cases excluding sequenced
    cases = base_store.cases(only_sequenced=True)

    # THEN cases should only contain the sequenced case
    assert cases
    for case in cases:
        assert new_case.internal_id in case.get("internal_id")


def test_only_delivered_cases(base_store: Store, helpers):
    """Test to that delivered cases can be included"""

    # GIVEN a database with an delivered case
    new_case = add_case(helpers, base_store)
    sample = helpers.add_sample(base_store, delivered_at=datetime.now())
    base_store.relate_sample(new_case, sample, "unknown")
    neg_new_case = add_case(helpers, base_store, "neg_new_case")
    neg_sample = helpers.add_sample(base_store, sample_name="neg_sample")
    base_store.relate_sample(neg_new_case, neg_sample, "unknown")

    # WHEN getting active cases excluding delivered
    cases = base_store.cases(only_delivered=True)

    # THEN cases should only contain the delivered case
    assert cases
    for case in cases:
        assert new_case.internal_id in case.get("internal_id")


def test_only_uploaded_cases(base_store: Store, helpers):
    """Test to that uploaded cases can be included"""

    # GIVEN a database with an uploaded analysis
    helpers.add_analysis(base_store, uploaded_at=datetime.now())
    neg_new_case = add_case(helpers, base_store, "neg_new_case")
    neg_sample = helpers.add_sample(base_store, sample_name="neg_sample")
    base_store.relate_sample(neg_new_case, neg_sample, "unknown")

    # WHEN getting active cases excluding uploaded
    cases = base_store.cases(only_uploaded=True)

    # THEN cases should only contain the uploaded case
    assert cases
    for case in cases:
        assert neg_new_case.internal_id not in case.get("internal_id")


def test_only_delivery_reported_cases(base_store: Store, helpers):
    """Test to that delivery-reported cases can be included"""

    # GIVEN a database with an delivery-reported analysis
    helpers.add_analysis(base_store, delivery_reported_at=datetime.now())
    neg_new_case = add_case(helpers, base_store, "neg_new_case")
    neg_sample = helpers.add_sample(base_store, sample_name="neg_sample")
    base_store.relate_sample(neg_new_case, neg_sample, "unknown")

    # WHEN getting active cases excluding delivery_reported
    cases = base_store.cases(only_delivery_reported=True)

    # THEN cases should only contain the delivery-reported case
    assert cases
    for case in cases:
        assert neg_new_case.internal_id not in case.get("internal_id")


def test_only_invoiced_cases(base_store: Store, helpers):
    """Test to that invoiced cases can be included"""

    # GIVEN a database with an invoiced case
    new_case = add_case(helpers, base_store)
    sample = helpers.add_sample(base_store)
    sample.invoice = base_store.add_invoice(helpers.ensure_customer(base_store))
    sample.invoice.invoiced_at = datetime.now()
    base_store.relate_sample(new_case, sample, "unknown")
    neg_new_case = add_case(helpers, base_store, "neg_new_case")
    neg_sample = helpers.add_sample(base_store, sample_name="neg_sample")
    base_store.relate_sample(neg_new_case, neg_sample, "unknown")

    # WHEN getting active cases excluding invoiced
    cases = base_store.cases(only_invoiced=True)

    # THEN cases should only contain the invoiced case
    assert cases
    for case in cases:
        assert new_case.internal_id in case.get("internal_id")


def test_only_analysed_cases(base_store: Store, helpers):
    """Test to that analysed cases can be included"""

    # GIVEN a database with a completed analysis
    helpers.add_analysis(base_store, completed_at=datetime.now())
    neg_new_case = add_case(helpers, base_store, "neg_new_case")
    neg_sample = helpers.add_sample(base_store, sample_name="neg_sample")
    base_store.relate_sample(neg_new_case, neg_sample, "unknown")

    # WHEN getting active cases excluding not analysed
    cases = base_store.cases(only_analysed=True)

    # THEN cases should only contain the completed case
    assert cases
    for case in cases:
        assert neg_new_case.internal_id not in case.get("internal_id")


def test_exclude_prepared_cases(base_store: Store, helpers):
    """Test to that sequenced cases can be excluded"""

    # GIVEN a database with an prepared case
    new_case = add_case(helpers, base_store)
    sample = helpers.add_sample(base_store, prepared_at=datetime.now())
    base_store.relate_sample(new_case, sample, "unknown")

    # WHEN getting active cases excluding prepared
    cases = base_store.cases(exclude_prepared=True)

    # THEN cases should not contain the prepared case
    assert not cases


def test_exclude_received_cases(base_store: Store, helpers):
    """Test to that sequenced cases can be excluded"""

    # GIVEN a database with an received case
    new_case = add_case(helpers, base_store)
    sample = helpers.add_sample(base_store, received_at=datetime.now())
    base_store.relate_sample(new_case, sample, "unknown")

    # WHEN getting active cases excluding received
    cases = base_store.cases(exclude_received=True)

    # THEN cases should not contain the received case
    assert not cases


def test_exclude_sequenced_cases(base_store: Store, helpers):
    """Test to that sequenced cases can be excluded"""

    # GIVEN a database with an sequenced case
    new_case = add_case(helpers, base_store)
    sample = helpers.add_sample(base_store, sequenced_at=datetime.now())
    base_store.relate_sample(new_case, sample, "unknown")

    # WHEN getting active cases excluding sequenced
    cases = base_store.cases(exclude_sequenced=True)

    # THEN cases should not contain the sequenced case
    assert not cases


def test_exclude_delivered_cases(base_store: Store, helpers):
    """Test to that delivered cases can be excluded"""

    # GIVEN a database with an delivered case
    new_case = add_case(helpers, base_store)
    sample = helpers.add_sample(base_store, delivered_at=datetime.now())
    base_store.relate_sample(new_case, sample, "unknown")

    # WHEN getting active cases excluding delivered
    cases = base_store.cases(exclude_delivered=True)

    # THEN cases should not contain the delivered case
    assert not cases


def test_exclude_uploaded_cases(base_store: Store, helpers):
    """Test to that uploaded cases can be excluded"""

    # GIVEN a database with an uploaded analysis
    helpers.add_analysis(base_store, uploaded_at=datetime.now())

    # WHEN getting active cases excluding uploaded
    cases = base_store.cases(exclude_uploaded=True)

    # THEN cases should not contain the uploaded case
    assert not cases


def test_exclude_delivery_reported_cases(base_store: Store, helpers):
    """Test to that delivery-reported cases can be excluded"""

    # GIVEN a database with an delivery-reported analysis
    helpers.add_analysis(base_store, delivery_reported_at=datetime.now())

    # WHEN getting active cases excluding delivery-reported
    cases = base_store.cases(exclude_delivery_reported=True)

    # THEN cases should not contain the delivery-reported case
    assert not cases


def test_exclude_invoiced_cases(base_store: Store, helpers):
    """Test to that invoiced cases can be excluded"""

    # GIVEN a database with an invoiced case
    new_case = add_case(helpers, base_store)
    sample = helpers.add_sample(base_store)
    sample.invoice = base_store.add_invoice(helpers.ensure_customer(base_store))
    sample.invoice.invoiced_at = datetime.now()
    base_store.relate_sample(new_case, sample, "unknown")

    # WHEN getting active cases excluding invoiced
    cases = base_store.cases(exclude_invoiced=True)

    # THEN cases should not contain the invoiced case
    assert not cases


def test_exclude_analysed_cases(base_store: Store, helpers):
    """Test to that analysed cases can be excluded"""

    # GIVEN a database with a completed analysis
    helpers.add_analysis(base_store, completed_at=datetime.now())

    # WHEN getting active cases excluding analysed
    cases = base_store.cases(exclude_analysed=True)

    # THEN cases should not contain the completed case
    assert not cases


def test_all_days(base_store: Store, helpers):
    """Test to that cases filter in case in database"""

    # GIVEN a database with a really old case
    new_case = add_case(helpers, base_store, ordered_days_ago=9999)

    # WHEN getting active cases with days = all
    cases = base_store.cases(days=0)

    # THEN cases should contain the case
    assert cases
    for case in cases:
        assert new_case.internal_id in case.get("internal_id")


def test_new_case_included(base_store: Store, helpers):
    """Test to that cases filter in case from database"""

    # GIVEN a database with a case
    new_case = add_case(helpers, base_store, ordered_days_ago=1)

    # WHEN getting active cases not older than two days
    result_cases = base_store.cases(days=2)

    # THEN cases should contain the case
    assert result_cases
    for case in result_cases:
        assert new_case.internal_id in case.get("internal_id")


def test_old_case_not_included(base_store: Store, helpers):
    """Test to that cases filter out old case in database"""

    # GIVEN a database with a case
    add_case(helpers, base_store, ordered_days_ago=2)

    # WHEN getting active cases not older than one day
    cases = base_store.cases(days=1)

    # THEN cases should not contain the case
    assert not cases


def test_analysis_bool_true(base_store: Store, helpers):
    """Test to that cases displays correct booleans for samples"""

    # GIVEN a database with a case
    helpers.add_analysis(base_store, completed_at=datetime.now(), uploaded_at=datetime.now())

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain true for all analysis booleans
    assert cases
    for case in cases:
        assert case.get("analysis_completed_bool") is True
        assert case.get("analysis_uploaded_bool") is True


def test_samples_bool_true(base_store: Store, helpers):
    """Test to that cases displays correct booleans for samples"""

    # GIVEN sample that is received, prepared, sequenced and delivered
    new_case = add_case(helpers, base_store)
    sample = helpers.add_sample(
        base_store,
        received_at=datetime.now(),
        prepared_at=datetime.now(),
        sequenced_at=datetime.now(),
        delivered_at=datetime.now(),
    )
    assert sample.received_at
    assert sample.prepared_at
    assert sample.sequenced_at
    assert sample.delivered_at
    sample.invoice = base_store.add_invoice(helpers.ensure_customer(base_store))
    sample.invoice.invoiced_at = datetime.now()
    base_store.relate_sample(new_case, sample, "unknown")

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain true for all sample booleans
    assert cases
    for case in cases:
        assert case.get("samples_received_bool") is True
        assert case.get("samples_prepared_bool") is True
        assert case.get("samples_sequenced_bool") is True
        assert case.get("samples_delivered_bool") is True
        assert case.get("samples_invoiced_bool") is True


def test_bool_false(base_store: Store, helpers):
    """Test to that cases displays correct received samples"""

    # GIVEN a database with a case
    new_case = add_case(helpers, base_store)
    sample = helpers.add_sample(base_store)
    base_store.relate_sample(new_case, sample, "unknown")

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain false for all booleans
    assert cases
    for case in cases:
        assert case.get("samples_received_bool") is False
        assert case.get("samples_prepared_bool") is False
        assert case.get("samples_sequenced_bool") is False
        assert case.get("analysis_completed_bool") is False
        assert case.get("analysis_uploaded_bool") is False
        assert case.get("samples_delivered_bool") is False
        assert case.get("samples_invoiced_bool") is False


def test_one_invoiced_sample(base_store: Store, helpers):
    """Test to that cases displays correct invoiced samples"""

    # GIVEN a database with a case with an invoiced sample
    new_case = add_case(helpers, base_store)
    sample = helpers.add_sample(base_store)

    sample.invoice = base_store.add_invoice(helpers.ensure_customer(base_store))
    sample.invoice.invoiced_at = datetime.now()

    base_store.relate_sample(new_case, sample, "unknown")
    assert sample.invoice is not None

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain number of invoiced samples
    assert cases
    for case in cases:
        assert case.get("samples_invoiced") == 1


def test_analysis_uploaded_at(base_store: Store, helpers):
    """Test to that cases displays when uploaded """

    # GIVEN a database with an analysis that was uploaded
    analysis = helpers.add_analysis(base_store, uploaded_at=datetime.now())
    assert analysis.uploaded_at is not None

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain info on upload occasion
    assert cases
    for case in cases:
        assert case.get("analysis_uploaded_at") is not None


def test_analysis_pipeline(base_store: Store, helpers):
    """Test to that cases displays pipeline """

    # GIVEN a database with an analysis that has pipeline
    pipeline = Pipeline.BALSAMIC
    analysis = helpers.add_analysis(base_store, pipeline=pipeline)
    assert analysis.pipeline is not None

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain info on pipeline
    assert cases
    for case in cases:
        assert case.get("analysis_pipeline") == str(pipeline)


def test_samples_delivered(base_store: Store, helpers):
    """Test to that cases displays when they were delivered """

    # GIVEN a database with a sample that is delivered
    new_case = add_case(helpers, base_store)
    sample = helpers.add_sample(base_store, delivered_at=datetime.now())
    base_store.relate_sample(new_case, sample, "unknown")
    assert sample.delivered_at is not None

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain info on delivered occasion
    assert cases
    for case in cases:
        assert case.get("samples_delivered") == 1


def test_analysis_completed_at(base_store: Store, helpers):
    """Test to that cases displays when they were completed """

    # GIVEN a database with an analysis that is completed
    analysis = helpers.add_analysis(base_store, completed_at=datetime.now())
    assert analysis.completed_at is not None
    assert base_store.families().count() == 1
    assert base_store.analyses().count() == 1

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain info on completion occasion
    assert cases
    for case in cases:
        assert case.get("analysis_completed_at") is not None


def test_case_ordered_date(base_store: Store, helpers):
    """Test to that cases displays when they were ordered """

    # GIVEN a database with a case without samples and no analyses
    new_case = add_case(helpers, base_store)
    assert new_case.ordered_at is not None

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain info on order occasion
    assert cases
    for case in cases:
        assert case.get("ordered_at") is not None


def test_one_of_two_samples_received(base_store: Store, helpers):
    """Test to that cases displays correct received samples"""

    # GIVEN a database with a case with a received sample and one not received
    new_case = add_case(helpers, base_store)
    sample_received = helpers.add_sample(base_store, "sample_received", received_at=datetime.now())
    sample_not_received = helpers.add_sample(base_store, "sample_not_received", received=False)
    base_store.relate_sample(new_case, sample_received, "unknown")
    base_store.relate_sample(new_case, sample_not_received, "unknown")
    assert sample_received.received_at is not None
    assert sample_not_received.received_at is None

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain number of received samples
    assert cases
    for case in cases:
        assert case.get("total_samples") == 2
        assert case.get("samples_received") == 1


def test_one_sequenced_sample(base_store: Store, helpers):
    """Test to that cases displays correct sequenced samples"""

    # GIVEN a database with a case with a sequenced sample
    new_case = add_case(helpers, base_store)
    sample = helpers.add_sample(base_store, sequenced_at=datetime.now())
    base_store.relate_sample(new_case, sample, "unknown")
    assert sample.sequenced_at is not None

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain number of sequenced samples
    assert cases
    for case in cases:
        assert case.get("samples_sequenced") == 1


def test_one_prepared_sample(base_store: Store, helpers):
    """Test to that cases displays correct prepared samples"""

    # GIVEN a database with a case
    new_case = add_case(helpers, base_store)
    sample = helpers.add_sample(base_store, prepared_at=datetime.now())
    base_store.relate_sample(new_case, sample, "unknown")
    assert sample.prepared_at is not None

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain number of prepared samples
    assert cases
    for case in cases:
        assert case.get("samples_prepared") == 1


def test_one_received_sample(base_store: Store, helpers):
    """Test to that cases displays correct received samples"""

    # GIVEN a database with a case
    new_case = add_case(helpers, base_store)
    sample = helpers.add_sample(base_store, received_at=datetime.now())
    base_store.relate_sample(new_case, sample, "unknown")
    assert sample.received_at is not None

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain number of received samples
    assert cases
    for case in cases:
        assert case.get("samples_received") == 1


def test_one_sample(base_store: Store, helpers):
    """Test to that cases displays correct total samples"""

    # GIVEN a database with a case
    new_case = add_case(helpers, base_store)
    sample = helpers.add_sample(base_store)
    base_store.relate_sample(new_case, sample, "unknown")
    assert sample.received_at is None
    assert sample.prepared_at is None
    assert sample.delivered_at is None
    assert sample.sequenced_at is None

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain the sample
    assert cases
    for case in cases:
        assert case.get("total_samples") == 1
        assert case.get("samples_received") == 0
        assert case.get("samples_prepared") == 0
        assert case.get("samples_sequenced") == 0
        assert case.get("samples_delivered") == 0
        assert case.get("samples_invoiced") == 0


def test_case_without_samples(base_store: Store, helpers):
    """Test to that cases displays correct number samples"""

    # GIVEN a database with a case without samples and no analyses
    new_case = add_case(helpers, base_store)
    assert not new_case.links
    assert not new_case.analyses

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain zero samples and no analysis info
    assert cases
    for case in cases:
        assert case.get("total_samples") == 0
        assert case.get("samples_received") is None
        assert case.get("samples_prepared") is None
        assert case.get("samples_sequenced") is None
        assert case.get("analysis_completed_at") is None
        assert case.get("analysis_pipeline") is None
        assert case.get("analysis_uploaded_at") is None
        assert case.get("samples_delivered") is None
        assert case.get("samples_invoiced") is None
        assert case.get("samples_received_bool") is None
        assert case.get("samples_prepared_bool") is None
        assert case.get("samples_sequenced_bool") is None
        assert case.get("analysis_completed_bool") is None
        assert case.get("analysis_uploaded_bool") is None
        assert case.get("samples_delivered_bool") is None
        assert case.get("samples_invoiced_bool") is None


def test_case_included(base_store: Store, helpers):
    """Test to that cases displays case in database"""

    # GIVEN a database with a case
    new_case = add_case(helpers, base_store)

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN cases should contain the case
    assert cases
    for case in cases:
        assert new_case.internal_id in case.get("internal_id")


def test_structure(base_store: Store, helpers):
    """Test to that cases displays case in database"""

    # GIVEN a database with a case
    add_case(helpers, base_store)

    # WHEN getting active cases
    cases = base_store.cases()

    # THEN structure should contain
    #   data_analysis
    #   internal_id
    #   name
    #   total number of samples,
    #   samples_received,
    #   samples_prepared,
    #   samples_sequenced,
    #   analysis_completed_at
    #   analysis_pipeline
    #   samples_delivered
    assert cases
    for case in cases:
        assert "data_analysis" in case.keys()
        assert "internal_id" in case.keys()
        assert "name" in case.keys()
        assert "ordered_at" in case.keys()
        assert "total_samples" in case.keys()
        assert "samples_received" in case.keys()
        assert "samples_prepared" in case.keys()
        assert "samples_sequenced" in case.keys()
        assert "analysis_completed_at" in case.keys()
        assert "analysis_pipeline" in case.keys()
        assert "samples_delivered" in case.keys()
        assert "samples_invoiced" in case.keys()


def add_case(
    helpers,
    disk_store,
    case_id="case_test",
    customer_id="cust_test",
    ordered_days_ago=0,
    action=None,
    priority=None,
    data_analysis=Pipeline.BALSAMIC,
    data_delivery=DataDelivery.SCOUT,
):
    """utility function to add a case to use in tests"""
    panel = helpers.ensure_panel(disk_store)
    customer = helpers.ensure_customer(disk_store, customer_id)
    case = disk_store.add_case(
        data_analysis=data_analysis, data_delivery=data_delivery, name=case_id, panels=panel.name
    )
    case.customer = customer
    case.ordered_at = datetime.now() - timedelta(days=ordered_days_ago)
    if action:
        case.action = action
    if priority:
        case.priority = priority
    disk_store.add_commit(case)
    return case



"""This script tests the cli methods to create prerequisites and start a mip-dna analysis"""
import logging

from cg.cli.workflow.mip_dna.base import start, start_available
from cg.constants import EXIT_SUCCESS, Pipeline
from cg.meta.workflow.prepare_fastq import PrepareFastqAPI


def test_dry(cli_runner, dna_mip_context):
    """Test mip dna start with --dry option"""

    # GIVEN a dna_mip_context

    # WHEN using dry running
    result = cli_runner.invoke(start_available, ["--dry-run"], obj=dna_mip_context)

    # THEN command should have accepted the option happily
    assert result.exit_code == EXIT_SUCCESS


def test_dna_case_included(cli_runner, caplog, dna_case, dna_mip_context, mocker):
    """Test mip dna start with a DNA case"""

    caplog.set_level(logging.INFO)

    # GIVEN spring decompression is needed
    mocker.patch.object(PrepareFastqAPI, "is_spring_decompression_needed")
    PrepareFastqAPI.is_spring_decompression_needed.return_value = True

    mocker.patch.object(PrepareFastqAPI, "start_spring_decompression")
    PrepareFastqAPI.start_spring_decompression.return_value = True

    # GIVEN a case that is ready for MIP DNA analysis
    #   -> has a sample that is sequenced and has an dna-application (non-wts)
    for link in dna_case.links:
        sample = link.sample
        assert sample.sequenced_at
        assert sample.application_version.application.analysis_type not in "wts"
    assert not dna_case.analyses

    # WHEN running command
    result = cli_runner.invoke(start_available, ["--dry-run"], obj=dna_mip_context)

    # THEN command should have printed the case id
    assert result.exit_code == EXIT_SUCCESS
    case_mentioned = False
    for _, level, message in caplog.record_tuples:
        if dna_case.internal_id in message:
            case_mentioned = True
            assert level == logging.INFO
    assert case_mentioned


def test_rna_case_excluded(cli_runner, caplog, dna_mip_context, rna_case, mocker):
    """Test mip dna start with a RNA case"""

    # GIVEN spring decompression is needed
    mocker.patch.object(PrepareFastqAPI, "is_spring_decompression_needed")
    PrepareFastqAPI.is_spring_decompression_needed.return_value = True

    mocker.patch.object(PrepareFastqAPI, "start_spring_decompression")
    PrepareFastqAPI.start_spring_decompression.return_value = True

    # GIVEN a case that is ready for MIP RNA analysis
    #   -> has a sample that is sequenced and has an rna-application (wts)

    assert rna_case.data_analysis == str(Pipeline.MIP_RNA)
    for link in rna_case.links:
        sample = link.sample
        assert sample.sequenced_at

    # WHEN running command
    result = cli_runner.invoke(start_available, ["--dry-run"], obj=dna_mip_context)

    # THEN command should not mention the rna-case
    assert result.exit_code == EXIT_SUCCESS

    case_mentioned = False
    for _, level, message in caplog.record_tuples:
        if rna_case.internal_id in message:
            case_mentioned = True
            assert level == logging.WARNING
    assert not case_mentioned


def test_mixed_dna_rna_case(cli_runner, caplog, dna_mip_context, dna_rna_mix_case, mocker):
    """Test mip dna start with a mixed DNA/RNA case"""

    # GIVEN spring decompression is needed
    mocker.patch.object(PrepareFastqAPI, "is_spring_decompression_needed")
    PrepareFastqAPI.is_spring_decompression_needed.return_value = True

    mocker.patch.object(PrepareFastqAPI, "start_spring_decompression")
    PrepareFastqAPI.start_spring_decompression.return_value = True

    # GIVEN a case that is ready for MIP RNA analysis
    #   -> has a sample that is sequenced and has an rna-application (wts)

    rna_sample_found = False
    non_rna_sample_found = False
    assert dna_rna_mix_case.data_analysis == str(Pipeline.MIP_DNA)
    for link in dna_rna_mix_case.links:
        sample = link.sample
        assert sample.sequenced_at
        if sample.application_version.application.analysis_type in "wts":
            rna_sample_found = True
        if sample.application_version.application.analysis_type not in "wts":
            non_rna_sample_found = True
    assert not dna_rna_mix_case.analyses
    assert rna_sample_found
    assert non_rna_sample_found

    # WHEN running command
    result = cli_runner.invoke(start_available, ["--dry-run"], obj=dna_mip_context)

    # THEN command should info about it starting the case but warn about skipping
    assert result.exit_code == EXIT_SUCCESS

    case_mentioned = False
    for _, level, message in caplog.record_tuples:
        if dna_rna_mix_case.internal_id in message:
            case_mentioned = True
            assert level == logging.WARNING
    assert case_mentioned



"""Tests for utils in store"""

import logging

from cg.store.models import FamilySample
from cg.store.utils import case_exists, get_links, get_samples


def test_case_exists_when_missing(case_id, caplog):
    """Test function when a case object is missing """
    caplog.set_level(logging.DEBUG)
    # GIVEN a case_id
    # WHEN case_obj is None
    was_found = case_exists(None, case_id)
    # THEN case_exists should return None
    assert was_found is None
    # THEN communicate case_id is missing
    assert f"{case_id}: case not found" in caplog.text


def test_case_exists_when_present(case_obj, case_id):
    """Test function when a case object is found """
    # GIVEN a case_id
    # WHEN case_obj is True
    was_found = case_exists(case_obj, case_id)
    # THEN case_exists should return True
    assert was_found


def test_get_samples(analysis_obj):
    """Test function to get all samples from a analysis object"""
    # GIVEN an analysis object
    # WHEN fetching all samples
    samples = get_samples(analysis_obj)
    # THEN assert the samples are FamilySamples
    for sample in samples:
        assert isinstance(sample, FamilySample)


def test_link(case_obj):
    """Test function to get all samples from a case"""
    # GIVEN an case object
    # WHEN fetching all links
    link_objs = get_links(case_obj)
    # THEN assert the link objs are samples
    for link_obj in link_objs:
        assert isinstance(link_obj, FamilySample)



"""Tests for the config part of Crunchy"""
import json

import pytest
from marshmallow import ValidationError

from cg.apps.crunchy import CrunchyAPI
from cg.apps.crunchy.models import CrunchyFileSchema


def test_map_fastq_file_info(file_schema):
    """Test to validate file information with the schema"""
    # GIVEN some correct file informtion and a file schema
    file_info = {
        "path": "a_file.fastq.gz",
        "file": "first_read",
        "checksum": "checksum_first_read",
        "algorithm": "sha256",
    }

    # WHEN validating the file information
    result = file_schema.load(file_info)

    # THEN assert that the validation was succesfull
    assert result
    # THEN assert that the object is correct
    assert result["path"] == file_info["path"]


def test_map_fastq_file_info_missing_path(file_schema):
    """Test to validate invalid file information with the schema"""
    # GIVEN some file informtion without a path and a file schema
    file_info = {
        "file": "first_read",
        "checksum": "checksum_first_read",
        "algorithm": "sha256",
    }

    # WHEN validating the file information
    with pytest.raises(ValidationError):
        # THEN assert that the validation was not succesfull
        file_schema.load(file_info)


def test_map_multiple_files(spring_metadata, file_schema):
    """Test to map a list of some files on the file schema"""
    # GIVEN a list of files
    assert len(spring_metadata) > 1
    result = []

    # WHEN validating the file information
    for file_info in spring_metadata:
        result.append(file_schema.load(file_info))

    # THEN assert that the validation was succesfull
    assert result
    # THEN assert that all files exists after dumping
    assert len(result) == len(spring_metadata)


def test_map_multiple_files_at_once(spring_metadata):
    """Test to map a list of some files on the file schema"""
    # GIVEN a list of files
    assert len(spring_metadata) > 1
    file_schema = CrunchyFileSchema(many=True)

    # WHEN validating the file information
    result = file_schema.load(spring_metadata)

    # THEN assert that the validation was succesfull
    assert isinstance(result, list)
    # THEN assert that all files exists after dumping
    assert len(result) == len(spring_metadata)


def test_get_spring_metadata_real_file(real_spring_metadata_path, crunchy_config_dict):
    """Test to parse the content of a real spring metadata file"""
    # GIVEN the path to a file with spring metadata content
    crunchy_api = CrunchyAPI(crunchy_config_dict)
    with open(real_spring_metadata_path, "r") as infile:
        content = json.load(infile)

    # WHEN parsing the content
    parsed_content = crunchy_api.get_spring_metadata(real_spring_metadata_path)

    # THEN assert the content is the same
    assert len(content) == len(parsed_content)


def test_update_date(spring_metadata_file, crunchy_config_dict):
    """Test to update the date in a spring metadata file"""
    # GIVEN the path to a metadata file without any "updated" information and a crunchy api
    crunchy_api = CrunchyAPI(crunchy_config_dict)
    spring_metadata = crunchy_api.get_spring_metadata(spring_metadata_file)
    for file_info in spring_metadata:
        assert "updated" not in file_info

    # WHEN running the update date function
    crunchy_api.update_metadata_date(spring_metadata_file)

    # THEN assert that the "updated" information has been added
    updated_spring_metadata = crunchy_api.get_spring_metadata(spring_metadata_file)
    for file_info in updated_spring_metadata:
        assert "updated" in file_info



from cg.store import Store, models
import datetime as dt
from alchy import Query


def get_cases_to_compress(store: Store, date_threshold: dt.datetime) -> Query:
    """Fetch all cases that are ready to be compressed by SPRING"""
    return (
        store.query(models.Family)
        .filter(models.Family.action not in ["running", "analyze"])
        .filter(models.Family.created_at < date_threshold)
        .order_by(models.Family.created_at.asc())
    )



"""Test methods for cg/cli/delete/cases"""
import logging

import pytest
from cg.cli.delete.cases import cases
from cg.store import Store

SUCCESS = 0


@pytest.mark.parametrize("identifier_key", ["ticket_number", "order"])
def test_set_cases_by_sample_identifiers(
    cli_runner, base_context, base_store: Store, identifier_key, helpers, caplog
):
    # GIVEN a database with a case with a sample
    sample_obj = helpers.add_sample(base_store)
    sample_obj.ticket_number = 123456
    sample_obj.order = "An order"
    case = helpers.add_case(base_store)
    helpers.add_relationship(base_store, sample=sample_obj, case=case)
    identifier_value = getattr(sample_obj, identifier_key)

    caplog.set_level(logging.INFO)

    # WHEN calling delete cases with valid sample identifiers
    cli_runner.invoke(
        cases,
        ["--sample-identifier", identifier_key, identifier_value],
        obj=base_context,
    )

    # THEN it should name the case to be changed
    assert case.internal_id in caplog.text
    assert case.name in caplog.text


def test_delete_cases_with_dry_run(cli_runner, base_context, base_store: Store, helpers, caplog):
    """Test that the delete cases will not delete the cases in dry-run mode """
    # GIVEN a database with a case
    case_obj = helpers.add_case(base_store)
    case_id = case_obj.internal_id
    sample = helpers.add_sample(base_store)
    helpers.add_relationship(store=base_store, case=case_obj, sample=sample)

    # WHEN deleting a case
    caplog.set_level(logging.DEBUG)
    cli_runner.invoke(
        cases,
        ["--sample-identifier", "name", sample.name, "--dry-run"],
        obj=base_context,
    )

    # THEN then it should not have been deleted
    assert "Cases (that will NOT be deleted due to --dry-run):" in caplog.text
    assert case_id in caplog.text



"""Base module for building bioinfo workflow bundles for linking in Housekeeper"""
import datetime as dt
import logging
from typing import List

from _io import TextIOWrapper
from housekeeper.store import models as hk_models

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import HK_TAGS, MICROSALT_TAGS, MIP_DNA_TAGS, MIP_RNA_TAGS, Pipeline
from cg.exc import (
    AnalysisDuplicationError,
    BundleAlreadyAddedError,
    MandatoryFilesMissing,
    PipelineUnknownError,
)
from cg.meta.store import mip as store_mip
from cg.store import Store, models
from cg.store.utils import reset_case_action

ANALYSIS_TYPE_TAGS = {
    "wgs": MIP_DNA_TAGS,
    "wes": MIP_DNA_TAGS,
    "wts": MIP_RNA_TAGS,
    "microsalt": MICROSALT_TAGS,
}
LOG = logging.getLogger(__name__)


def gather_files_and_bundle_in_housekeeper(
    config_stream: TextIOWrapper, hk_api: HousekeeperAPI, status: Store, workflow: Pipeline
) -> models.Analysis:
    """Function to gather files and bundle in housekeeper"""

    add_analysis = {
        Pipeline.MIP_DNA: store_mip.add_mip_analysis,
    }
    bundle_data = add_analysis[workflow](config_stream)

    results = hk_api.add_bundle(bundle_data)
    if results is None:
        raise BundleAlreadyAddedError("bundle already added")

    bundle_obj, version_obj = results

    case_obj = {
        Pipeline.MIP_DNA: status.family(bundle_obj.name),
    }

    reset_case_action(case_obj[workflow])
    new_analysis = add_new_analysis(bundle_data, case_obj[workflow], status, version_obj)
    version_date = version_obj.created_at.date()

    LOG.info("new bundle added: %s, version %s", bundle_obj.name, version_date)
    hk_api.include(version_obj)
    hk_api.add_commit(bundle_obj, version_obj)

    return new_analysis


def build_bundle(config_data: dict, analysisinfo_data: dict, deliverables: dict) -> dict:
    """Create a new bundle to store in Housekeeper"""

    analysis_type = config_data["samples"][0]["type"]

    return {
        "name": config_data["case"],
        "created": analysisinfo_data["date"],
        "pipeline_version": analysisinfo_data["version"],
        "files": deliverables_files(deliverables, analysis_type),
    }


def add_new_analysis(
    bundle_data: dict,
    case_obj: models.Family,
    status: Store,
    version_obj: hk_models.Version,
) -> models.Analysis:
    """Function to create and return a new analysis database record"""

    try:
        pipeline = Pipeline(case_obj.data_analysis)
    except ValueError:
        raise PipelineUnknownError(f"No pipeline specified in {case_obj}")

    if status.analysis(family=case_obj, started_at=version_obj.created_at):
        raise AnalysisDuplicationError(
            f"Analysis object already exists for {case_obj.internal_id} {version_obj.created_at}"
        )

    new_analysis = status.add_analysis(
        pipeline=pipeline,
        version=bundle_data.get("pipeline_version"),
        started_at=version_obj.created_at,
        completed_at=dt.datetime.now(),
        primary=(len(case_obj.analyses) == 0),
    )
    new_analysis.family = case_obj
    return new_analysis


def deliverables_files(deliverables: dict, analysis_type: str) -> list:
    """Get all deliverable files from the pipeline"""
    pipeline_tags = HK_TAGS[str(analysis_type)]
    analysis_type_tags = ANALYSIS_TYPE_TAGS[str(analysis_type)]
    files = parse_files(deliverables, pipeline_tags, analysis_type_tags)
    _check_mandatory_tags(files, analysis_type_tags)
    return files


def _is_deliverables_tags_missing_in_analysis_tags(
    analysis_type_tags: dict, deliverables_tag_map: tuple
) -> bool:
    """Check if deliverable tags are represented in analysis tags """
    return deliverables_tag_map not in analysis_type_tags


def parse_files(deliverables: dict, pipeline_tags: list, analysis_type_tags: dict) -> list:
    """ Get all files and their tags from the deliverables file """

    parsed_files = []
    for file_ in deliverables["files"]:
        deliverables_tag_map = (
            (file_["step"],) if file_["tag"] is None else (file_["step"], file_["tag"])
        )
        if _is_deliverables_tags_missing_in_analysis_tags(
            analysis_type_tags=analysis_type_tags, deliverables_tag_map=deliverables_tag_map
        ):
            continue
        parsed_file = {
            "path": file_["path"],
            "tags": get_tags(file_, pipeline_tags, analysis_type_tags, deliverables_tag_map),
            "archive": False,
            "deliverables_tag_map": deliverables_tag_map,
        }
        parsed_files.append(parsed_file)

        if file_["path_index"]:
            parsed_index_file = {
                "path": file_["path_index"],
                "tags": get_tags(
                    file_, pipeline_tags, analysis_type_tags, deliverables_tag_map, is_index=True
                ),
                "archive": False,
                "deliverables_tag_map": deliverables_tag_map,
            }
            parsed_files.append(parsed_index_file)

    return parsed_files


def get_tags(
    file: dict,
    pipeline_tags: list,
    analysis_type_tags: dict,
    deliverables_tag_map: tuple,
    is_index: bool = False,
) -> List[str]:
    """Get all tags for a file"""

    tags = {
        "id": str(file["id"]),
        "pipeline": pipeline_tags[0],
        "application": pipeline_tags[1] if len(pipeline_tags) > 1 else None,
    }

    return _convert_deliverables_tags_to_hk_tags(
        tags, analysis_type_tags, deliverables_tag_map, is_index
    )


def _convert_deliverables_tags_to_hk_tags(
    tags: dict, analysis_type_tags: dict, deliverables_tag_map: tuple, is_index: bool = False
) -> List[str]:
    """
    Filter and convert tags from external deliverables tags to standard internal housekeeper
    tags
    """

    if is_index:
        mapped_tags = analysis_type_tags[deliverables_tag_map]["index_tags"]
    else:
        mapped_tags = analysis_type_tags[deliverables_tag_map]["tags"]
    converted_tags = [tags["id"], tags["pipeline"]]

    if tags["application"] is not None:
        converted_tags.append(tags["application"])

    for tag in mapped_tags:
        converted_tags.append(tag)

    return sorted(list(set(converted_tags)))


def _check_mandatory_tags(files: list, pipeline_tags: dict):
    """
    Check if all the mandatory tags are present for the files to be added to Housekeeper.
    Raise an exception if not.
    """
    deliverable_tags = [file_["deliverables_tag_map"] for file_ in files]
    mandatory_tags = [tag for tag in pipeline_tags if pipeline_tags[tag]["is_mandatory"]]

    tags_are_missing, missing_tags = _determine_missing_tags(mandatory_tags, deliverable_tags)

    if tags_are_missing:
        raise MandatoryFilesMissing(
            f"Mandatory files are missing! These are the missing tags: {missing_tags}."
        )


def _determine_missing_tags(mandatory_tags: list, found_tags: list) -> tuple:
    """
    Determines if mandatory tags are missing and hence if files are missing in the
    deliverables, and returns any missing tags
    """

    missing_tags = set(mandatory_tags) - set(found_tags)
    are_tags_missing = bool(len(missing_tags) > 0)

    return are_tags_missing, missing_tags



"""Fixtures for the shipping api tests"""

from pathlib import Path

from tests.apps.shipping.conftest import (
    fixture_shipping_configs,
    fixture_binary_path,
    fixture_host_config,
)

import pytest


@pytest.fixture(name="scout_config")
def fixture_scout_config(apps_dir: Path) -> Path:
    """Return the path to a shipping config for scout"""
    return apps_dir / "shipping" / "scout-deploy.yaml"
